# MT::ObjectDriver::Redland
# -----------------
# $Revision: 1.5 $
# $Date: 2005/04/03 22:16:21 $
# ---------------------------------------------------------

=head1 NAME

MT::ObjectDriver::Redland - MoveableType driver to add a Redland storage mechanism.

=cut

package MT::ObjectDriver::Redland;

=head1 SYNOPSIS

Once installed in your MovableType lib/MT/ObjectDriver directory,
the standard database configuration directives can be used in mt.cfg:

 ObjectDriver Redland
 Database greg
 DBUser greg
 DBHost localhost

The Redland driver currently only uses the Redland mysql storage type. The
database password should be stored in mt-db-pass.cgi, as with any other
database driver.

=head1 REQUIRES

 MovableType 3.1 or higher
 RDF::Redland
 Sort::Naturally
 Encode

=head1 DESCRIPTION

Blah blah blah.

=cut

use strict;
use warnings;
use Carp qw(carp confess);

use Encode;
use RDF::Redland;
use Data::Dumper;
use Sort::Naturally;

use MT::Util qw( offset_time_list );
use base qw(MT::ObjectDriver);

our ($VERSION, $REVISION, $debug, $blankseq);
BEGIN {
	$debug				= 0;
	$blankseq			= 1;
	$REVISION			= do { my @REV = split(/\./, (qw$Revision: 1.5 $)[1]); sprintf("%0.3f", $REV[0] + ($REV[1]/1000)) };
	$VERSION			= 0.002;
}

sub init {
	my $driver = shift;
	$driver->SUPER::init(@_);
	my $cfg = $driver->cfg;

	my $db		= $cfg->Database;
	my $host	= $cfg->DBHost;
	my $user	= $cfg->DBUser;
	my $pass	= $cfg->DBPassword;
	my @args	= ("mysql", 'mt', qq(host='localhost',database='${db}',user='${user}',password='${pass}'));
	my $storage	= new RDF::Redland::Storage(@args);
	my $model	= new RDF::Redland::Model($storage, "");
	$driver->{model} = $model or return $driver->error(MT->translate("Connection error: [_1]", ''));	# XXX
	$driver;
}

sub model {
	my $driver	= shift;
	return $driver->{model};
}

sub load {
	my $driver = shift;
	carp "load ( @_ )" if ($debug);
	my $iter	= $driver->load_iter( @_ );
	my @data;
	while (my $obj = $iter->()) {
		return $obj unless wantarray;
		push(@data, $obj);
	}
	return @data;
}

sub load_iter {
	my $driver = shift;
	$driver->run_callbacks($_[0] . '::pre_load', \@_);
	carp "load_iter ( @_ )" if ($debug);
	my ($class, $terms, $args) = @_;
	$args		||= {};
	$terms		= {} unless (defined($terms));
	my $urin	= $class->get_rdf_class_node;
	my $uri		= $urin->uri->as_string;
	my $model	= $driver->{model};
	
	my $nodes;
	if (my $join = $args->{'join'}) {
		my ($jclass, $jcol, $jterms, $jargs)	= @{ $join };
		my $iter	= $jclass->load_iter( $jterms, $jargs );
		my $pred	= $class->get_column_rdf_node( 'id' );
		my $const	= $class->get_rdf_column_constructor( 'id' );
		$args->{'unique'}	||= $jargs->{'unique'} if (ref($jargs));
		$nodes		= sub {
			my $jnode	= $iter->();
			return undef unless $jnode;
			my $col	= $jnode->$jcol();
			my $val	= RDF::Redland::Node->$const( $col );
			my ($node)	= $model->sources( $pred, $val );
			return $node;
		};
	} else {
		my $rdql		= $driver->_prepare_rdql($class, $terms, $args);
		my $query		= RDF::Redland::Query->new($rdql);
		my $i			= $model->query_execute($query);
		$nodes		= sub {
			return undef if (not $i or not ref($i));
			return undef if ($i->finished);
			my $node	= $i->binding_value(0);
			$i->next_result;
			return $node;
		};
	}
	
	my %unique;
	my $unique		= $args->{'unique'};
	my $limit		= $args->{'limit'};
	my $finished	= 0;
	my $iter		= sub {
		my $node;
		while (1) {
			return undef if ($finished);
			if ($limit) {
				$limit--;
				$finished	= 1 if ($limit == 0);
			}
			$node	= $nodes->();
			unless ($node) {
				$finished	= 1;
				return undef;
			}
			
			if ($unique) {
				warn $node->as_string;
				next if ($unique{ $node->as_string }++);
			}
			last;
		}
		my %data;
		my @cols		= @{ $class->column_names };
		my %datecols	= map { $_ => 1 } qw(created_on modified_on);
		foreach my $col (@cols) {
			my $pred	= $class->get_column_rdf_node( $col );
			my (@val)	= $model->targets( $node, $pred );
			if (@val) {
				my $val	= $val[0];
				if ($val->is_resource) {
					$data{$col}	= $val->uri->as_string;
				} elsif ($val->is_literal) {
					$data{$col}	= decode('utf8', $val->literal_value);
				} else {
					$data{$col}	= $val->blank_identifier;
				}
				$data{$col}	= $driver->db2ts($data{$col}) if ($datecols{ $col });
			}
		}
		my $obj 	= $class->new;
		$obj->set_values(\%data);
		$driver->run_callbacks($class . '::post_load', $terms, $args, $obj);
		warn Dumper($obj) if ($debug > 1);
		return $obj;
	};
	
	if (exists $args->{'sort'}) {
		my $col	= $args->{'sort'};
		my $asc	= (exists $args->{'direction'}) ? ($args->{'direction'} eq 'ascend') : 1;
		my @objs;
		while (my $obj = $iter->()) {
			push(@objs, $obj);
		}
		@objs	= map { $_->[0] } sort { ncmp($a->[1], $b->[1]) } map { [$_, $_->$col()] } @objs;
		@objs	= reverse @objs unless ($asc);
		$iter	= sub {
			return shift(@objs);
		};
	}
	
	return $iter;
}

sub insert {
	my $driver	= shift;
	carp "insert ( @_ )" if ($debug);
	my $obj		= shift;
	my $cols	= $obj->column_names;
	unless ($obj->id) {
		my $id = $driver->generate_id($obj);
		$obj->id($id);
	}
	if ($obj->properties->{audit}) {
		my $blog_id	= $obj->blog_id;
		my @ts	= offset_time_list(time, $blog_id);
		my $ts	= sprintf '%04d-%02d-%02dT%02d:%02d:%02dZ', $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
		$obj->created_on($ts) unless $obj->created_on;
		$obj->modified_on($ts);
	}
	
	my $model	= $driver->{model};
	my $node	= $obj->get_rdf_node( $model );
	my $uri		= $obj->get_rdf_type_node;
	my $type	= RDF::Redland::Node->new_from_uri('http://www.w3.org/1999/02/22-rdf-syntax-ns#type');
	my $st		= RDF::Redland::Statement->new( $node, $type, $uri );
	if (not $model->contains_statement( $st )) {
		$model->add_statement( $st );
	}
	
	my $col_defs	= $obj->properties->{column_defs};
	my %datecols	= map { $_ => 1 } qw(created_on modified_on);
	foreach my $col (@{ $cols }) {
		my $val		= $obj->column($col) || '';
		$val		= $driver->ts2db($val, $obj, $col) if ($datecols{ $col });
		my $pred	= $obj->get_column_rdf_node( $col );
		my $method	= $obj->get_rdf_column_constructor( $col );
		my $obj		= RDF::Redland::Node->$method( $val );
		my $st		= RDF::Redland::Statement->new( $node, $pred, $obj );
		$driver->{model}->add_statement( $st );
	}
	1;
}

sub update {
	my $driver	= shift;
	carp "update ( @_ )" if ($debug);
	my $obj		= shift;
	my $cols	= $obj->column_names;
	if ($obj->properties->{audit}) {
		my $blog_id	= $obj->blog_id;
		my @ts	= offset_time_list(time, $blog_id);
		my $ts	= sprintf '%04d-%02d-%02dT%02d:%02d:%02dZ', $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
		$obj->created_on($ts) unless $obj->created_on;
		$obj->modified_on($ts);
	}
	
	my $model	= $driver->{model};
	my $node	= $obj->get_rdf_node( $model );
	my %datecols	= map { $_ => 1 } qw(created_on modified_on);
	foreach my $col (@{ $cols }) {
		next unless (exists $obj->{'column_values'}{$col});
		my $val		= $obj->column($col) || '';
		$val		= $driver->ts2db($val, $obj, $col) if ($datecols{ $col });
		my $pred	= $obj->get_column_rdf_node( $col );
		my $method	= $obj->get_rdf_column_constructor( $col );
		my $st		= RDF::Redland::Statement->new( $node, $pred, undef );
		my @st		= $model->find_statements( $st );
		my $new		= RDF::Redland::Node->$method($val);
		my $add		= 1;
		foreach my $old (@st) {
			if ($old->object->equals( $new )) {
				$add	= 0;
			} else {
				$model->remove_statement( $old );
			}
		}
		if ($add) {
			$st->object( $new );
			$model->add_statement( $st );
		}
	}
	1;
}

sub save {
	my $driver	= shift;
	carp "save ( @_ )" if ($debug);
	my $obj		= shift;
    my $class	= ref($obj);
    (my $original, $obj)	= ($obj, $obj->clone());
	$driver->run_callbacks($class . '::pre_save', $obj, $original);
	my $result;
	if ($driver->exists($obj)) {
		$result	= $driver->update($obj);
	} else {
		$result	= $driver->insert($obj);
	}
	$original->id($obj->id);
	$original->created_on($obj->created_on);
	$original->modified_on($obj->modified_on);
	$driver->run_callbacks($class . '::post_save', $obj);
	return $result;
}

sub count {
	my $driver = shift;
	carp "count ( @_ )" if ($debug);
	my $class	= shift;
	my $iter	= $class->load_iter( @_ );
	my $count	= 0;
	while ($iter->()) {
		$count++;
	}
	return $count;
}

sub exists {
	my $driver = shift;
	carp "exists ( @_ )" if ($debug);
	my $obj		= shift;
    return unless $obj->id;
	my $model	= $driver->{model};
	my $pred	= $obj->get_column_rdf_node( 'id' );
	my $val		= RDF::Redland::Node->new_literal($obj->id);
	my @nodes	= $model->sources( $pred, $val );
	return scalar(@nodes);
}

sub remove {
	my $driver = shift;
	carp "remove ( @_ )" if ($debug);
	my $obj		= shift;
	my $class	= ref($obj);
	$driver->run_callbacks($class . '::pre_remove', $obj);
	my $id		= $obj->id;
	return $driver->error("No such object.") unless defined($id);
	my $model	= $driver->{model};
	my $pred	= $class->get_column_rdf_node( 'id' );
	my @nodes	= $model->sources( $pred, RDF::Redland::Node->new_literal($id) );
	foreach my $node (@nodes) {
		my $st		= RDF::Redland::Statement->new( $node, undef, undef );
		my (@st)	= $model->find_statements( $st );
		$model->remove_statement($_) for (@st);
	}
	$driver->run_callbacks($class . '::post_remove', $obj);
	1;
}

sub remove_all {
	my $driver = shift;
	carp "remove_all ( @_ )" if ($debug);
	my $proto	= shift;
	my $class	= ref($proto) || $proto;
	$driver->run_callbacks($class . '::pre_remove_all');
	my $model	= $driver->{model};
	my $pred	= RDF::Redland::Node->new_from_uri('http://www.w3.org/1999/02/22-rdf-syntax-ns#type');
	my $type	= $class->get_rdf_type_node;
	my @nodes	= $model->sources( $pred, $type );
	foreach my $node (@nodes) {
		my $st		= RDF::Redland::Statement->new( $node, undef, undef );
		my (@st)	= $model->find_statements( $st );
		$model->remove_statement($_) for (@st);
	}
	$driver->run_callbacks($class . '::post_remove_all');
	1;
}

sub generate_id {
	my $driver	= shift;
	my $class	= shift;
	if (ref($driver) and $class->can('get_rdf_class_node')) {
		my $node	= $class->get_rdf_class_node;
		my $model	= $driver->{model};
		my $pred	= RDF::Redland::Node->new_from_uri('http://kasei.us/e/ns/mt#seq_next');
		my ($stp)	= RDF::Redland::Statement->new( $node, $pred, undef );
		my ($st)	= $model->find_statements( $stp );
		my $val		= 1;
		if ($st) {
			$model->remove_statement($st);
			$val	= $st->object->literal_value;
		}
		$stp->object( RDF::Redland::Node->new_literal( $val+1 ) );
		$model->add_statement($stp);
		return $val;
	} else {
		our $blankseq;
		return sprintf("%s%04d", time, $blankseq++);
	}
}

sub ts2db {
	my $driver	= shift;
	my $date	= shift;
	my $obj		= shift;
	my $col		= shift;
	my $tz		= 'Z';
	if ($obj and $col ne 'modified_on') {
		if (my $blog = $obj->blog) {
			my $offset	= $blog->server_offset;
			my $partial_hour_offset = 60 * abs($offset - int($offset));
			$tz	= sprintf("%s%02d:%02d", $offset < 0 ? '-' : '+', abs($offset), $partial_hour_offset);
		}
	}
	
	if ($date =~ m#^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$#) {
		$date		= "$1-$2-$3T$4:$5:$6${tz}";
	}
	return $date;
}

sub db2ts {
	my $driver	= shift;
	my $date	= shift;
	$date		=~ s/[^0-9]//g;
	$date		=~ s/^(\d{14}).*$/$1/;
	return $date;
}

sub replace_node {
	my $model		= shift;
	my $node		= shift;
	my $url			= shift;
	my $replacewith	= RDF::Redland::Node->new_from_uri( $url );
	return if ($node->equals( $replacewith ));
	my $sameas		= RDF::Redland::Node->new_from_uri( 'http://www.w3.org/2002/07/owl#sameAs' );
	
	foreach ([$node, $replacewith], [$replacewith, $node]) {
		my $equivst		= RDF::Redland::Statement->new( $_->[0], $sameas, $_->[1] );
		unless ($model->contains_statement( $equivst )) {
			$model->add_statement( $equivst );
		}
	}
	warn "Replacing " . $node->as_string . " with " . $replacewith->as_string . "\n" if ($debug);
	foreach my $data (
						[RDF::Redland::Statement->new( $node, undef, undef ), 'subject'],
						[RDF::Redland::Statement->new( undef, undef, $node ), 'object']
					) {
		my ($stmt, $method)	= @{ $data };
		my (@stmts)	= $model->find_statements( $stmt );
		foreach my $st (@stmts) {
			next if ($st->predicate->as_string eq '[http://www.w3.org/2002/07/owl#sameAs]');	# don't touch equivalences
			my $to	= $st->clone;
			$to->$method( $replacewith );
			if ($model->contains_statement( $to )) {
				warn "Model already contains statement: " . $to->as_string if ($debug);
			} else {
				$model->add_statement( $to );
				$model->remove_statement( $st );
				if ($debug > 1) {
					warn "-- \n";
					warn $st->as_string . "\n";
					warn $to->as_string . "\n";
				}
			}
		}
	}
}

sub _prepare_rdql {
	my $driver	= shift;
	my ($class, $terms, $args)	= @_;
	Carp::cluck( Dumper($class, $terms, $args) ) if ($debug > 1);
	my $typen	= $class->get_rdf_type_node;
	my $type	= $typen->uri->as_string;
	my @where;
	push(@where, "(?node rdf:type <${type}>)");
	my $source	= $class->datasource;
	$terms	= { "id" => $terms } if ($terms and not ref($terms));
	while (my($col,$val) = each(%{ $terms })) {
		no warnings 'uninitialized';
		$val		= qq("$val");
		my $predn	= $class->get_column_rdf_node( $col );
		my $pred	= $predn->uri->as_string;
		push(@where, qq#(?node <${pred}> ${val})#);
	}
	
	my $where	= join("\n\t", @where);
	my $rdql	= "SELECT\n\t?node\nWHERE\n\t${where}";
	warn $rdql if ($debug > 1);
	return $rdql;
}

######################################################################
######################################################################
###  Shove these methods into existing MT::Object classes:

package MT::Object;
	sub get_rdf_type_node {
		&get_rdf_class_node;
	}
	sub get_rdf_class_node {
		my $class	= shift;
		my $name	= $class->datasource;
		my $uri		= "http://kasei.us/e/ns/mt/${name}";
		return RDF::Redland::Node->new_from_uri( $uri );
	}
	sub get_rdf_node {
		my $self	= shift;
		my $model	= shift || $self->driver->model;
		if (my $id = $self->id) {
			my $pred	= $self->get_column_rdf_node( 'id' );
			my $val		= RDF::Redland::Node->new_literal($id);
			my (@nodes)	= $model->sources( $pred, $val );
			return $nodes[0] if (@nodes);
		}
		my $node	= RDF::Redland::Node->new_from_blank_identifier( MT::ObjectDriver::Redland->generate_id( $self ) );
		return $node;
	}
	sub get_column_rdf_node {
		my $self	= shift;
		my $col		= shift;
		my %map		= $self->get_column_rdf_node_map;
		my $class	= $self->get_rdf_class_node->uri->as_string;
		my $uri		= (exists $map{ $col }) ? $map{ $col } : join('#', $class, $col);
		return RDF::Redland::Node->new_from_uri( $uri );
	}
	sub get_column_rdf_node_map {
		return ();
	}
	sub get_rdf_column_constructor {
		my $self	= shift;
		my $col		= shift;
		my %map		= (
						url		=> 'new_from_uri',
					);
		return $map{ $col } || 'new_literal';
	}

package MT::Blog;
	sub get_rdf_column_constructor {
		my $self	= shift;
		my $col		= shift;
		my %map		= (
						site_url	=> 'new_from_uri',
						archive_url	=> 'new_from_uri'
					);
		return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
	}

package MT::Entry;
	sub get_rdf_type_node {
		return RDF::Redland::Node->new_from_uri('http://purl.org/rss/1.0/item')
	}
	sub get_rdf_node {
		my $self	= shift;
		my $model	= $self->driver->model;
		my $url		= $self->archive_url;
		if (my $id = $self->id) {
			my $pred	= $self->get_column_rdf_node( 'id' );
			my $val		= RDF::Redland::Node->new_literal($id);
			my ($node)	= $model->sources( $pred, $val );
			if ($node) {
				if ($url ne $node->uri->as_string) {
					MT::ObjectDriver::Redland::replace_node( $model, $node, $url );
				}
				return $node;
			}
		}
		my $node	= RDF::Redland::Node->new_from_uri( $url );
		return $node;
	}
	sub get_column_rdf_node_map {
		return (
				created_on	=> 'http://purl.org/dc/terms/created',
				modified_on	=> 'http://purl.org/dc/terms/modified'
			);
	}

package MT::TBPing;
	sub get_rdf_column_constructor {
		my $self	= shift;
		my $col		= shift;
		my %map		= (
						source_url		=> 'new_from_uri',
					);
		return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
	}

package MT::Author;
	sub get_rdf_type_node {
		return RDF::Redland::Node->new_from_uri('http://xmlns.com/foaf/0.1/Person')
	}
	sub get_column_rdf_node_map {
		return (
				name		=> 'http://xmlns.com/foaf/0.1/name',
				nickname	=> 'http://xmlns.com/foaf/0.1/nick',
				email		=> 'http://xmlns.com/foaf/0.1/mbox',
				mboxsha		=> 'http://xmlns.com/foaf/0.1/mbox_sha1sum',
				url			=> 'http://xmlns.com/foaf/0.1/homepage'
			);
	}
	sub get_rdf_column_constructor {
		my $self	= shift;
		my $col		= shift;
		my %map		= (
						email		=> 'new_from_uri',
						url			=> 'new_from_uri'
					);
		return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
	}
	sub column_names {
		my $self	= shift;
		my $names	= $self->SUPER::column_names;
		push(@{ $names }, qw(mboxsha));
		return $names;
	}

######################################################################

use MT::Plugin;
use Digest::SHA1 qw(sha1_hex);
use Data::Dumper::Simple;

my $plugin	= new MT::Plugin({	name		=> "Redland RDF Plugin, v1.000",
								description	=> "Adds an RDF storage mechanism" });

MT::Author->add_callback("post_load", 1, $plugin, sub {
	my ($cb, $terms, $args, $obj) = @_;
	my $email	= $obj->email;
	$email		=~ s/^mailto://;
	$obj->email( $email );
});

MT::Author->add_callback("pre_save", 1, $plugin, sub {
	my ($cb, $obj)	= @_;
	if (my $email = $obj->email) {
		$email	= "mailto:${email}";
		$obj->email( $email );
		my $sha	= sha1_hex( $email );
		$obj->mboxsha( $sha );
	}
});

1;

__END__

=head1 REVISION HISTORY

 $Log: Redland.pm,v $
 Revision 1.5  2005/04/03 22:16:21  greg
 - updated VERSION number

 Revision 1.4  2005/04/02 22:51:12  greg
 - updated the doc_link MT plugin variable to point to MT-Redland
 - CMS hooks can now add multiple boxes instead of all being in the same box

 Revision 1.3  2005/04/02 22:37:58  greg
 - implemented join, unique, limit, and sort in load_iter()

 Revision 1.2  2005/04/01 21:19:51  greg
 - Updating an entry with a new URL now does the Right Thing
   (including inserting owl:sameAs triples)


=head1 COPYRIGHT

Copyright (c) 2005 Gregory Williams. All rights reserved. This program is free
software; you can redistribute it and/or modify it under the same terms as Perl
itself.

=head1 AUTHOR

Gregory Williams <greg@evilfunhouse.com>

=cut
