# MT::Plugins::Redland::References
# -----------------
# $Revision: 1.3 $
# $Date: 2005/04/02 22:51:12 $
# ---------------------------------------------------------

=head1 NAME

MT::Plugins::Redland::References - Adds dcterms:references statements to the Redland model

=cut

package MT::Plugins::Redland::References;

=head1 SYNOPSIS

New triples will be added to the Redland model using dcterms:references
for every hyperlink present in the body of an entry.

=head1 REQUIRES

 MovableType 2.6 or higher
 HTML::Parser

=head1 DESCRIPTION

This plugin uses the dcterms:references predicate to link an
entry's RDF resource with those it references in HTML. See
L<http://web.resource.org/rss/1.0/modules/dcterms/#references>. 

All HTML anchors containing an href attribute are parsed from the
current entry's text, and added as dcterms:references statements.
For example, given the HTML:

  I like <a href="http://www.megatokyo.com/">MegaTokyo</a>.

the following RDF statement will be added to the model:

  {[http://example.com/ilike.html], [http://purl.org/dc/terms/references], [http://www.megatokyo.com/]}

=cut

use strict;
use warnings;

use IO::All;
use Data::Dumper;
use RDF::Redland;

use MT 2.6;
use MT::Template::Context;

BEGIN {
	no warnings 'numeric';
	our $NAME		= "Redland Reviews";
	our $VERSION	= '0.001';
	our $plugin_data	= {
			name		=> "${NAME}, v${VERSION}",
			description => 'Adds review statements to the Redland model',
			doc_link	=> 'http://kasei.us/code/mtredland/'
		};
	
	if ($MT::VERSION >= 3.0) {
		require MT::Plugin;
		MT::Plugin->import();
		our $plugin	= new MT::Plugin($plugin_data);
		MT->add_plugin($plugin);
	}
}

our $debug		= 0;
use Data::Dumper;
use HTML::Parser;

######################################################################

our $plugin;
my $cfg		= MT::ConfigMgr->instance;
if ($cfg->ObjectDriver eq 'Redland') {
	MT::Entry->add_callback("pre_save", 5, $plugin, sub {
		my ($cb, $entry)	= @_;
		my $html	= $entry->text;
		my $model	= $entry->driver->model;
		if ($html =~ m#Rating:\s*(\d+)/(\d+)#smi) {
			my $rating	= RDF::Redland::Node->new_literal( int( 10 * $1 / $2 ) );
			my $node	= $entry->get_rdf_node;
			my $author	= $entry->author->get_rdf_node;
			my $ns		= 'http://www.purl.org/stuff/rev#';
			my $p_type		= RDF::Redland::Node->new_from_uri( "http://www.w3.org/1999/02/22-rdf-syntax-ns#type" );
			my $p_has		= RDF::Redland::Node->new_from_uri( "${ns}hasReview" );
			my $p_rev		= RDF::Redland::Node->new_from_uri( "${ns}Review" );
			my $p_rating	= RDF::Redland::Node->new_from_uri( "${ns}rating" );
			my $p_reviewer	= RDF::Redland::Node->new_from_uri( "${ns}reviewer" );
			
			my $st		= RDF::Redland::Statement->new( $node, $p_has, undef );
			my ($rev)	= $model->find_statements( $st );
			unless ($rev) {
				$rev	= RDF::Redland::Node->new_from_blank_identifier('r' . time . 'r' . $MT::ObjectDriver::Redland::blankseq++);
				$model->add_statement( RDF::Redland::Statement->new( $rev, $p_type, $p_rev ) );
				$model->add_statement( RDF::Redland::Statement->new( $node, $p_has, $rev ) );
			}
			
			my @revs	= $model->targets( $rev, $p_reviewer );
			$model->remove_statement( $_ ) for (grep { not $author->equals( $_ ) } @revs);
			my @ratings	= $model->targets( $rev, $p_rating );
			$model->remove_statement( $_ ) for (grep { not $rating->equals( $_ ) } @ratings);
			
			$model->add_statement( RDF::Redland::Statement->new( $rev, $p_reviewer, $author ) );
			$model->add_statement( RDF::Redland::Statement->new( $rev, $p_rating, $rating ) );
		}
	});
}

	
######################################################################

1;

__END__

=head1 SEE ALSO

 L<MT::Plugin>
 L<HTML::Parser>
 L<http://web.resource.org/rss/1.0/modules/dcterms/#references>

=head1 KNOWN BUGS

None.

=head1 TODO

None.

=head1 LICENSE

Copyright (c) 2005 Gregory Williams. All rights reserved. This program is free
software; you can redistribute it and/or modify it under the same terms as Perl
itself.

=head1 AUTHOR

 Gregory Williams <gwilliams@cpan.org>

=cut
